/****************************************************************************
*
*   Copyright (c) 2012 Pau Pajuelo
*
*   This program is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License version 2 as
*   published by the Free Software Foundation.
*
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with this program; if not, write to the Free Software
*   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
*   MA 02110-1301 USA.
*
****************************************************************************
*
*   Send numbers via UART (RS232)
*   GPIO Debug option needs user-gpio-drv.ko. Go to:
*   http://labs.igep.es/index.php/How_to_use_GPIOs
*
****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <asm/ioctls.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>

//Constants and structure defined in GPIO functions
#define GPIO_CMD_DIRECTION_OUTPUT       0x83
#define GPIO_MAGIC  'G'
#define GPIO_IOCTL_DIRECTION_OUTPUT     _IOW(   GPIO_MAGIC, GPIO_CMD_DIRECTION_OUTPUT, GPIO_Value_t )

typedef struct
{
    unsigned    gpio;
    int         value;

} GPIO_Value_t;

//Define Array char size
#define SIZE 30
//Define Debug mode:0==Disable
#define DEBUG 0

//Define wait select
#define WAIT 5

//Set GPIO Debugger port
//Be sure that mux is correctly configured and GPIOs are not used
#define GPIO1 136
#define GPIO2 137

//Initializes the GPIO library.
//Error<0 and Not Error==0
int gpio_init(int *gFd )
{
    if ( (*gFd) < 0 )
    {
        if (( (*gFd) = open( "/dev/user-gpio", O_RDWR )) < 0 )
        {
            return -1;
        }
    }
    return 0;
}

//Terminates the GPIO library.
void gpio_term( int gFd )
{
    if ( gFd>=0 )
    {
        close(gFd);
    }
}

//Configures a GPIO for output and sets the initial value.
//Error<0 and Not Error==0
int  gpio_direction_output( unsigned gpio, int initialValue ,int gFd)
{
    GPIO_Value_t    setDirOutput;

    setDirOutput.gpio = gpio;
    setDirOutput.value = initialValue;

    if ( ioctl( gFd, GPIO_IOCTL_DIRECTION_OUTPUT, &setDirOutput ) != 0 )
    {
        return -errno;
    }
    return 0;
}

//Open Uart Port
//Error<0 and Not Error==0
int open_uart(char *uart_name)
{
    int fd;

    //Open flags:
    //O_RDWR: Open fd like Read/Write mode
    //O_NDELAY: When possible, the file is opened in non-blocking mode.
    //O_NOCTTY: Tells UNIX that this program doesn't want to be the "controlling terminal" for that port
    fd=open(uart_name, O_RDWR | O_NDELAY | O_NOCTTY);
    if (fd>=0)
        return fd;

    return -1;
}

//Initialize some program parts, is necessary pass ti pointer point to NULL
//This program allocate memory is important to free with end_comm
//Error<0 and Not Error==0
int init_comm(struct termios **ti, int speedint, int fd)
{
    speed_t speed;
    struct termios *fti;

    if ((*ti)!=NULL)
    {
        return -2;
    }

    (*ti)=malloc(sizeof(struct termios));
    fti=(*ti);
    if (fti==NULL)
    {
        return -3;
    }

    //Clean buffers
    tcflush(fd,TCIFLUSH);
    tcsetattr(fd, TCSANOW, (*ti));

    //Choose speed:
    switch (speedint)
    {
        case 300:
            speed|=B300;
            break;
        case 600:
            speed|=B600;
            break;
        case 1200:
            speed|=B1200;
            break;
        case 2400:
            speed|=B2400;
            break;
        case 4800:
            speed|=B4800;
            break;
        case 9600:
            speed|=B9600;
            break;
        case 19200:
            speed|=B19200;
            break;
        case 38400:
            speed|=B38400;
            break;
        case 57600:
            speed|=B57600;
            break;
        case 115200:
            speed|=B115200;
            break;
        case 230400:
            speed|=B230400;
            break;
        case 460800:
            speed|=B460800;
            break;
        case 921600:
            speed|=B921600;
            break;
        case 3000000:
            speed|=B3000000;
            break;
        default:
            return -1;
            break;
    }

    //Get current device atributes
    tcgetattr(fd, fti);
    //Set baud rates
    cfsetispeed(fti, speed);
    cfsetospeed(fti, speed);

    //Set other options:

    //CS8: Transmit 8 bits
    //CLOCAL: Ignore modem control lines
    //CREAD:  Enable receiver
    //IGNPAR:  Ignore framing errors and parity errors
    //ICRNL: Translate  carriage  return to newline on input
    //VMIN: Minimum number of characters for non-canonical read
    //VTIME: Timeout in deciseconds for non-canonical read.
    //ICANON: Enable canonical mode:
    //Note: every time that send '\n' select lets read uart buffer

/*   Canonical mode: Input is made available line by line.  An  input  line is
     available when one of the line delimiters is typed (NL, EOL, EOL2; or EOF
     at the start of line).  Except in the case of EOF, the line delimiter  is
     included in the buffer returned by read(2).

     Line  editing  is  enabled  (ERASE, KILL;  and if the IEXTEN flag is set:
     WERASE,  REPRINT, LNEXT).  A  read(2)  returns  at  most  one  line  of
     input; if  the  read(2) requested  fewer bytes than are available in the
     current line of input, then  only as many  bytes as  requested are  read,
     and  the remaining characters will be available  for a  future  read(2).*/
     //More information: man termios

    //Set control flags
    fti->c_cflag |= CS8 | CLOCAL | CREAD;
    //Set input flags
    fti->c_iflag |= IGNPAR | ICRNL;
    //Set output flags (no flags configured)
    fti->c_oflag |= 0;
    //Set local flags
    fti->c_lflag |= ICANON;

    //Don't wait between receivin characters
    fti->c_cc[VMIN]=1;
    fti->c_cc[VTIME]=0;

    //Clean buffers
    tcflush(fd, TCIFLUSH);

    //Set new atributes
    tcsetattr(fd, TCSANOW,fti);

    return 0;
}

//Clean and close some program parts
void end_comm(struct termios **ti, int fd)
{
    tcflush(fd,TCIFLUSH);
    tcsetattr(fd, TCSANOW, (*ti));
    close(fd);
    free((*ti));
    ti=NULL;
}

//MAIN PROGRAM
//Error<0 and Not Error==0
int main (int argc, char *argv[])
{
    char device[SIZE], speedstr[SIZE],numberstr[SIZE],numberrcv[SIZE];
    int opt,fd,res,speedint,number, oldnumber=-1, percentageint=0,debug_count=DEBUG,Gpio_debug=0,gFd=-1;
    struct termios *ti=NULL;
    float percentage;
    fd_set ready;
    struct timeval tv;

    //Initialize arrays of chars
    memset(device, '\0',SIZE);
    memset(speedstr, '\0',SIZE);
    memset(numberstr, '\0',SIZE);
    memset(numberrcv, '\0',SIZE);

    //getopt part:
    while ((opt = getopt(argc, argv, "d:s:n:D::")) != -1)
    {
        switch (opt)
        {
            case 'd':
            snprintf(device,SIZE,"%s",optarg);
            break;
            case 's':
            snprintf(speedstr,SIZE,"%s",optarg);
            speedint=atoi(speedstr);
            break;
            case 'n':
            snprintf(numberstr,SIZE,"%s",optarg);
            break;
            case 'D':
            Gpio_debug=1;
            break;
            default: // ’?’
            fprintf(stderr, "Usage: %s \n",argv[0]);
            return -1;
        }
    }

    if (optind < argc)
    {
        printf("Non-option ARGV-elements: ");
        while (optind < argc)
            printf("%s ", argv[optind++]);

        printf("\n");
    }

    if (device[0]!='\0' && speedstr[0]!='\0' && numberstr[0]!='\0')
    {
        //open uart:
        fd=open_uart(device);
        if (fd<0)
        {
            printf("Error: open_uart\n");
            return -1;
        }
        //open gpio:
        if (Gpio_debug)
        {
            if ( gpio_init(&gFd) < 0 )
            {
                printf( "gpio_init failed\n");
                return -1;
            }
        }

        //Parse number
        res=atoi(numberstr);
        if (res<0)
        {
            printf("Error: number won't be negative\n");
            if (Gpio_debug)
                gpio_term(gFd);
            return -1;
        }

        //Initialize uart:
        res=init_comm(&ti,speedint,fd);
        if(res<0)
        {
            printf("Error init_comm: ");
            if (res==-1)
            {
                printf("baud rate must be 300, 600, 1200, 2400, 4800, 9600, 19200, 38400, 57600, 115200, 230400, 460800, 921600 or 3000000\n");
            }
            if (res==-2)
            {
                printf("input var ti punter don't have NULL value\n");
            }
            if (res==-3)
            {
                printf("problem with malloc\n");
            }
            if (Gpio_debug)
                gpio_term(gFd);
            return -1;

        }

        //Prepare first value to send
        strcat(numberstr,"\n");

        while (1)
        {
            //Uart debug option
            #if DEBUG
            if(!(debug_count--))
            {
                printf("Exit by debug\n");
                end_comm;
                if (Gpio_debug)
                    gpio_term(gFd);
                return 0;
            }
            #endif

            //write part:
            printf("Send value:\t\t%s", numberstr);
            if (Gpio_debug)
            {
                //GPIO1 High value
                if( gpio_direction_output( GPIO1, 1,gFd) < 0 )
                {
                    printf( "gpio_direction_output GPIO1 failed value HIGH\n" );
                    gpio_term(gFd);
                    end_comm;
                    return -1;
                }
            }
            res=write(fd,numberstr,strlen(numberstr));

            if (Gpio_debug)
            {
                //GPIO1 Low value
                if( gpio_direction_output( GPIO1, 0,gFd) < 0 )
                {
                    printf( "gpio_direction_output GPIO1 failed value LOW\n" );
                    gpio_term(gFd);
                    end_comm;
                    return -1;
                }
            }
            if (res<0)
            {
                printf("Error write\n");
                end_comm;
                if (Gpio_debug)
                    gpio_term(gFd);
                return -1;
            }

            //Prepare select variable ready:
            //Put ready bits to 0
            FD_ZERO (&ready);
            //Put fd value to ready
            FD_SET(fd,&ready);

            //Prepare select variable tv
            tv.tv_sec=WAIT;
            tv.tv_usec=0;

            //select part:
            if (Gpio_debug)
            {
                if( gpio_direction_output( GPIO2, 1,gFd) < 0 )
                {
                    //GPIO2 High value
                    printf( "gpio_direction_output GPIO2 failed value HIGH\n" );
                    gpio_term(gFd);
                    end_comm;
                    return -1;
                }
            }
            res=select(fd+1,&ready, NULL, NULL,&tv);

            if (Gpio_debug)
            {
                if( gpio_direction_output( GPIO2, 0,gFd) < 0 )
                {
                    //GPIO2 Low value
                    printf( "gpio_direction_output GPIO2 failed value LOW\n" );
                    gpio_term(gFd);
                    end_comm;
                    return -1;
                }
            }
            switch (res)
            {
                case 0:
                    printf("select timeout\n\n");
                    break;
                case -1:
                    printf("select error\n");
                    end_comm;
                    return -1;

                default:
                    //FD_ISSET: Returns a non-zero value if the file descriptor is set in the file descriptor set pointed to by fdset; otherwise returns 0.
                    if (FD_ISSET(fd, &ready))
                    {
                        //read part:
                        memset(numberrcv, '\0',SIZE);
                        res=read(fd,numberrcv,SIZE);
                        switch (res)
                        {
                            case 0:
                            case -1:
                                printf("Error read\n");
                                end_comm;
                                if (Gpio_debug)
                                    gpio_term(gFd);
                                return -1;
                            default:
                                //Change '\n' for '\0'
                                numberrcv[strlen(numberrcv)]='\0';
                                printf("Avaliable with read:\t%s\n",numberrcv);
                                number=atoi(numberrcv);
                                //Parse data valid
                                percentageint++;
                                if (!((oldnumber<=-1) || (number==(oldnumber+2))))
                                {
                                    percentage=(100.0/percentageint);
                                    printf("Data information corrupted, error(100): %1.6f\n",percentage);
                                    end_comm;
                                    if (Gpio_debug)
                                        gpio_term(gFd);
                                    return -1;
                                }
                                //Increase number and write part:
                                oldnumber=number;
                                number++;
                                memset(numberstr, '\0',SIZE);
                                snprintf(numberstr,SIZE,"%d\n",number);
                                break;
                        }
                    }
            }
        }
        end_comm;
        if (Gpio_debug)
            gpio_term(gFd);
        return 0;
    }
    printf("\n\t\t Uart Tutorial Program\n");
	printf("Use:\n\n");
	printf("\tExecutable -d<option> -n<option> -s<option> -D\n");
	printf("\t\t-d  :Device directory\n");
	printf("\t\t-s  [300 .. 3000000]: Bus speed\n");
    printf("\t\t-n  [integer]: Number send, will be positive\n");
    printf("\t\t-D  :GPIO Debug option (optional)\n\n");
    return -1;
}
